#!/usr/bin/python

import sys
import os
import subprocess
import time
from produce_graphs import draw_graph
bin_dir_path = os.path.dirname(os.path.realpath(__file__))

OUTPUT_DIR=time.strftime("%Y_%m_%d_%H_%M_%S")

#Parameters
THREAD_COUNTS="1,2,3,4,5,6,7,8"#1,2,4,8,16,32,64
PINNING="no"
NR_OF_WARM_UP_RUNS_INSERT_DELETE="10"
NR_OF_WARM_UP_RUNS="1"
WARM_UP_TIME_SECONDS="9"
NR_OF_MEASURMENT_RUNS="3"
MEASURMENT_TIME_SECONDS="3"
REPORT_TYPE='AVERAGE_THROUGHPUT' #  ALL,AVERAGE_MIN_MAX,AVERAGE or AVERAGE_THROUGHPUT


GENERIC_CONFIG_STR = (NR_OF_WARM_UP_RUNS + "@" +
                      WARM_UP_TIME_SECONDS + "@" +
                      NR_OF_MEASURMENT_RUNS + "@" +
                      MEASURMENT_TIME_SECONDS + "@" +
                      REPORT_TYPE)
#,se.uu.collection.CATreeMapOpt
DATA_STRUCTURES='se.uu.collection.CATreeMap,se.uu.collection.CATreeMapSafeOpt,java.util.concurrent.ConcurrentSkipListMap'#,edu.stanford.ppl.concurrent.SnapTreeMap,contention.datastructures.lockBased.LockBasedFriendlyTreeMap,trees.logicalordering.LogicalOrderingAVL

os.mkdir(OUTPUT_DIR)

def insert_bench_command(key_range_size   = 10000000,
                         nr_of_insert_ops = 1000000):
    return [
        os.path.join(bin_dir_path, 'benchmark_runner.py'),
        #Output dir (standard means a dir in bench_result based on the git
        #commit id and the date)
        OUTPUT_DIR, 
        #benchmark name
        'se.uu.bench.SetInsertBench',
        #Data structures (comma separated list)
        DATA_STRUCTURES,
        #Benchmark number of threads (comma separated list)
        THREAD_COUNTS,
        #use pinning to the first NUMA node
        PINNING,
        #Number of warm up runs
        NR_OF_WARM_UP_RUNS_INSERT_DELETE,
        #Number of measurments runs
        NR_OF_MEASURMENT_RUNS,
        #Result report type
        REPORT_TYPE,
        #Key range size
        str(key_range_size),
        #Nr of operations
        str(nr_of_insert_ops)]

def delete_bench_command(key_range_size   = 10000000,
                         nr_of_delete_ops = 1000000):
    return [
        os.path.join(bin_dir_path, 'benchmark_runner.py'),
        #Output dir (standard means a dir in bench_result based on the git
        #commit id and the date)
        OUTPUT_DIR, 
        #benchmark name
        'se.uu.bench.SetDeleteBench',
        #Data structures (comma separated list)
        DATA_STRUCTURES,
        #Benchmark number of threads (comma separated list)
        THREAD_COUNTS,
        #use pinning to the first NUMA node
        PINNING,
        #Number of warm up runs
        NR_OF_WARM_UP_RUNS_INSERT_DELETE,
        #Number of measurments runs
        NR_OF_MEASURMENT_RUNS,
        #Result report type
        REPORT_TYPE,
        #Key range size
        str(key_range_size),
        #Nr of operations
        str(nr_of_delete_ops)]


def mix_ops_bench_command(key_range_size    = 10000000,
                          nr_of_prefill_ops = 1000000,
                          percentage_reads  = 0.8):
    return [
        os.path.join(bin_dir_path, 'benchmark_runner.py'),
        #Output dir (standard means a dir in bench_result based on the git
        #commit id and the date)
        OUTPUT_DIR, 
        #benchmark name
        'se.uu.bench.SetMixBench',
        #Data structures (comma separated list)
        DATA_STRUCTURES,
        #Benchmark number of threads (comma separated list)
        THREAD_COUNTS,
        #use pinning to the first NUMA node
        PINNING,
        #Number of warm up runs
        NR_OF_WARM_UP_RUNS,
        #Seconds to run a warm up
        WARM_UP_TIME_SECONDS,
        #Number of measurments runs
        NR_OF_MEASURMENT_RUNS,
        #Seconds to run a measurment
        MEASURMENT_TIME_SECONDS,
        #Result report type
        REPORT_TYPE,
        #Key range size
        str(key_range_size),
        #Prefill ops
        str(nr_of_prefill_ops),
        #Persentage reads
        str(percentage_reads)]


def local_mix_ops_bench_command(key_range_size    = 10000000,
                                nr_of_prefill_ops = 1000000,
                                percentage_reads  = 0.8,
                                next_key_range_size  = 200):
    return [
        os.path.join(bin_dir_path, 'benchmark_runner.py'),
        #Output dir (standard means a dir in bench_result based on the git
        #commit id and the date)
        OUTPUT_DIR, 
        #benchmark name
        'se.uu.bench.SetLocalMixBench',
        #Data structures (comma separated list)
        DATA_STRUCTURES,
        #Benchmark number of threads (comma separated list)
        THREAD_COUNTS,
        #use pinning to the first NUMA node
        PINNING,
        #Number of warm up runs
        NR_OF_WARM_UP_RUNS,
        #Seconds to run a warm up
        WARM_UP_TIME_SECONDS,
        #Number of measurments runs
        NR_OF_MEASURMENT_RUNS,
        #Seconds to run a measurment
        MEASURMENT_TIME_SECONDS,
        #Result report type
        REPORT_TYPE,
        #Key range size
        str(key_range_size),
        #Prefill ops
        str(nr_of_prefill_ops),
        #Persentage reads
        str(percentage_reads),
        #Decides how much the key can change from the previous operation
        str(next_key_range_size)]


def run_command(command):
    process = subprocess.Popen(command)
    process.wait()

def run_mix_command_produce_graph(key_range_size    = 2000000,
                                  nr_of_prefill_ops = 1000000,
                                  percentage_reads  = 0.8,
                                  graph_producer_file=None):
    run_command(mix_ops_bench_command(key_range_size,
                                      nr_of_prefill_ops,
                                      percentage_reads))
    graph_producer_file.write('draw_graph("mix_' + str(percentage_reads) + "_" + str(key_range_size) + "_" + str(nr_of_prefill_ops) + '",\n')
    graph_producer_file.write('           "' + ("se.uu.bench.SetMixBench_" +
                                               GENERIC_CONFIG_STR + "@" +
                                               str(key_range_size) + "@" +
                                               str(nr_of_prefill_ops) + "@" +
                                               str(percentage_reads)) + '",\n')
    graph_producer_file.write('           xmax=None,\n')
    graph_producer_file.write('           ymax=None,\n')
    graph_producer_file.write('           legendLoc=0)\n')
    graph_producer_file.flush()

def run_local_mix_command_produce_graph(key_range_size    = 2000000,
                                        nr_of_prefill_ops = 1000000,
                                        percentage_reads  = 0.8,
                                        next_key_range_size  = 200,
                                        graph_producer_file=None):
    run_command(local_mix_ops_bench_command(key_range_size,
                                            nr_of_prefill_ops,
                                            percentage_reads,
                                            next_key_range_size))
    graph_producer_file.write('draw_graph("local_mix_' + str(percentage_reads) + "_" + str(next_key_range_size) + "_" + str(key_range_size) + "_" + str(nr_of_prefill_ops) + '",\n')
    graph_producer_file.write('           "' + ("se.uu.bench.SetLocalMixBench_" +
                                                GENERIC_CONFIG_STR + "@" +
                                                str(key_range_size) + "@" +
                                                str(nr_of_prefill_ops) + "@" +
                                                str(percentage_reads) + "@" +
                                                str(next_key_range_size)) + '",\n')
    graph_producer_file.write('           xmax=None,\n')
    graph_producer_file.write('           ymax=None,\n')
    graph_producer_file.write('           legendLoc=0)\n')
    graph_producer_file.flush()

    
#RUN BENCHMARKS

# run_command(insert_bench_command(key_range_size   = 10000000,
#                                  nr_of_insert_ops = 1000000))

# run_command(delete_bench_command(key_range_size   = 10000000,
#                                  nr_of_delete_ops = 1000000))



with open(OUTPUT_DIR + "/graph_producer_file.py", "a", 1) as graph_producer_file:
    graph_producer_file.write('#!/usr/bin/python\n')
    graph_producer_file.write('import sys\n')
    graph_producer_file.write('sys.path.append("..")\n')    
    graph_producer_file.write('from produce_graphs import draw_graph\n')
    for size in [1000000,1000]:#10000000,,100000,10000,1000,100,10
        for percentage_reads_param in [0.99,0.7,0.0]:#1.0,0.99,0.9,0.7,0.5,0.0
            run_mix_command_produce_graph(key_range_size    = size*2,
                                          nr_of_prefill_ops = size,
                                          percentage_reads  = percentage_reads_param,
                                          graph_producer_file = graph_producer_file)
            run_local_mix_command_produce_graph(size * 2,
                                                size,
                                                percentage_reads  = percentage_reads_param,
                                                next_key_range_size  = 10,
                                                graph_producer_file = graph_producer_file)


